"""Module for Message base definition."""
import re

from pydantic_avro.base import AvroBase  # type: ignore


class Message(AvroBase):  # type: ignore
    """Base model for events sent/received by clients."""

    def get_key(self) -> bytes:
        """Returns the message key to used by Kafka."""
        raise NotImplementedError("Please implement the method get_key")

    @classmethod
    def get_topic_name(cls) -> str:
        """Returns the topic name to be used for this message."""
        name = cls.__name__
        snake_case_name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()
        return snake_case_name



"""Module for the base settings definition."""
from pydantic import BaseModel
from pydantic import BaseSettings


class MessageBrokerSettings(BaseModel):
    """Representation of the settings available for the event bus."""

    broker_host: str = "localhost:9092"
    schema_registry_host: str = "http://localhost:8081"


class Settings(BaseSettings):
    """Representation for the SDK settings."""

    message_broker: MessageBrokerSettings = MessageBrokerSettings()

    class Config:
        """Defines how the this model is configured."""

        env_file = ".env"
        env_file_encoding = "utf-8"
        env_nested_delimiter = "__"



"""Module for classes for creating Kafka Sink Connectors."""
from typing import Dict
from typing import Union

import requests
from pydantic import BaseModel

ConnectorConfig = Dict[str, Union[str, bool, int]]


class BaseConnector(BaseModel):
    """Base class for all Kafka connector configurations."""

    name: str
    topics: str
    key_converter: str = "org.apache.kafka.connect.storage.StringConverter"
    transforms: str = "removenullfields"
    transforms_removenullfields_type: str = (
        "us.anant.kafka.connect.smt.RemoveNullFields$Value"
    )

    def _generate_config(self) -> ConnectorConfig:
        config: ConnectorConfig = {}
        for k, v in self.dict().items():
            if k == "name":
                continue
            new_k = k.replace("_", ".")
            config[new_k] = v
        return config

    def create_connector(self) -> None:
        """Creates the Kafka Connector.

        If it already exists updates the configuration.
        """
        url = f"http://localhost:8083/connectors/{self.name}/config"
        config = self._generate_config()
        headers = {"content-type": "application/json"}
        res = requests.put(url=url, json=config, headers=headers)

        if res.status_code > 201:
            raise ValueError(
                f"""
                Failed to create connector with code {res.status_code}.
                Details: {res.json()}
                """
            )


class ElasticConnector(BaseConnector):
    """Configuration for Elasticsearch connector.

    More details at
    https://docs.confluent.io/kafka-connect-elasticsearch/current/configuration_options.html#elasticsearch-overview-config
    """

    username: str
    password: str
    connection_url: str
    connector_class: str = (
        "io.confluent.connect.elasticsearch.ElasticsearchSinkConnector"
    )
    type_name: str = "_doc"
    schema_ignore: bool = False
    key_ignore: bool = False
    behavior_on_null_values: str = "delete"
    behavior_on_malformed_documents: str = "ignore"
    write_method: str = "upsert"


class PostgresConnector(BaseConnector):
    """Configuration for Postgres connector.

    More details at
    https://docs.confluent.io/kafka-connect-jdbc/current/sink-connector/sink_config_options.html#sink-config-options
    """

    connection_user: str
    connection_password: str
    connection_url: str
    pk_fields: str
    connector_class: str = "io.confluent.connect.jdbc.JdbcSinkConnector"
    # connector_class: str = (
    #     "no.norsktipping.kafka.connect.jdbc.connector.JdbcSinkConnector_Flatten"
    # )
    auto_create: bool = True
    auto_evolve: bool = True
    pk_mode: str = "record_key"
    insert_mode: str = "upsert"
    delete_enabled: bool = True
    transforms: str = "dropsome,removenullfields"
    transforms_dropsome_blacklist: str = ""
    transforms_dropsome_type: str = (
        "org.apache.kafka.connect.transforms.ReplaceField$Value"
    )


class MongoConnector(BaseConnector):
    """Configuration for MongoDB connector.

    More details at
    https://docs.mongodb.com/kafka-connector/current/sink-connector/configuration-properties/#std-label-kafka-sink-configuration-properties
    """

    database: str
    collection: str
    connection_uri: str
    connector_class: str = "com.mongodb.kafka.connect.MongoSinkConnector"
    mongodb_delete_on_null_values: bool = True
    delete_on_null_values: bool = True
    document_id_strategy_overwrite_existing: bool = True
    document_id_strategy: str = (
        "com.mongodb.kafka.connect.sink.processor.id.strategy.ProvidedInKeyStrategy"
    )
    transforms: str = "hk,removenullfields"
    transforms_hk_type: str = "org.apache.kafka.connect.transforms.HoistField$Key"
    transforms_hk_field: str = "_id"
    writemodel_strategy: str = (
        "com.mongodb.kafka.connect.sink.writemodel.strategy.UpdateOneTimestampsStrategy"
    )



"""Module for the Message Broker abstraction."""
from typing import Generator
from typing import Type
from typing import TypeVar

from kafka.admin import KafkaAdminClient  # type: ignore
from kafka.admin import NewTopic
from kafka.consumer import KafkaConsumer  # type: ignore
from kafka.producer import KafkaProducer  # type: ignore
from schema_registry.client import schema  # type: ignore
from schema_registry.client import SchemaRegistryClient
from schema_registry.serializers import AvroMessageSerializer  # type: ignore

from driva_python_sdk.core.settings import Settings
from driva_python_sdk.messages.message import Message


MessageType = TypeVar("MessageType", bound=Message)


class MessageBroker:
    """This is an abstraction to interact with the Kafka Message Broker."""

    def __init__(self) -> None:
        """Initializes the MessageBroker.

        It uses the settings automatically loaded from the environment.
        """
        self.setting = Settings()
        broker_host = self.setting.message_broker.broker_host
        self.kafka_admin_client = KafkaAdminClient(
            bootstrap_servers=[broker_host],
            client_id="admin_client",
        )
        self.kafka_producer = KafkaProducer(bootstrap_servers=[broker_host])
        schema_registry_url = self.setting.message_broker.schema_registry_host
        self.schema_registry_client = SchemaRegistryClient(url=schema_registry_url)
        self.message_serializer = AvroMessageSerializer(self.schema_registry_client)

    def create_topic_with_schema(
        self, cls: Type[Message], num_partitions: int = 1
    ) -> int:
        """Creates a topic with the given schema associated.

        The topic name will be infered from the class name.
        The schema name will be <topic>-value.
        """
        topic_name = cls.get_topic_name()
        subject_name = topic_name + "-value"
        avro_schema = cls.avro_schema()

        schema_id: int = self.schema_registry_client.register(
            subject=subject_name, schema=schema.AvroSchema(avro_schema)
        )
        topic = NewTopic(
            name=topic_name, num_partitions=num_partitions, replication_factor=1
        )
        self.kafka_admin_client.create_topics(new_topics=[topic])

        return schema_id

    def send(self, message: Message) -> None:
        """Send the message to the right topic.

        The name of the topic is infered from the class name.
        """
        topic_name = message.get_topic_name()
        subject_name = topic_name + "-value"
        avro_schema = message.avro_schema()

        key = message.get_key()
        value = self.message_serializer.encode_record_with_schema(
            subject=subject_name,
            schema=schema.AvroSchema(avro_schema),
            record=message.dict(exclude_none=True, exclude_unset=True),
        )
        self.kafka_producer.send(topic=topic_name, value=value, key=key)
        self.kafka_producer.flush()

    def consume(
        self, cls: Type[MessageType], from_beggining: bool = True
    ) -> Generator[MessageType, None, None]:
        """Consumes messages from the right topic.

        The name of the topic is infered from the class name.
        Raises ValueError if can't deserialize the message.
        """
        topic_name = cls.get_topic_name()
        consumer = KafkaConsumer(
            topic_name, auto_offset_reset="earliest" if from_beggining else "latest"
        )
        for msg in consumer:
            value = self.message_serializer.decode_message(msg.value)
            if value is None:
                raise ValueError("Failed to deserialize message")
            message = cls(**value)
            yield message
